@@ -1,11 +1,15 @@
 from __future__ import annotations
 
+import errno
+import fcntl
 import os
+import select
 import sys
 import tempfile
 import textwrap
 from functools import partial
 from pathlib import Path
+from typing import Any
 from typing import Callable
 from typing import Dict
 from typing import List
@@ -17,6 +21,8 @@ from PyQt5.QtCore import QThreadPool
 from PyQt5.QtCore import Qt
 from PyQt5.QtCore import pyqtSignal
 from PyQt5.QtCore import pyqtSlot
+from PyQt5.QtGui import QDoubleValidator
+from PyQt5.QtGui import QIntValidator
 from PyQt5.QtGui import QPainter
 from PyQt5.QtGui import QTextCursor
 from PyQt5.QtSvg import QSvgRenderer
@@ -83,6 +89,22 @@ class FunctionThreadSignals(QObject):
     data = pyqtSignal(object)
 
 
+# Helper function to add the O_NONBLOCK flag to a file descriptor
+def make_async(fd):
+    fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)
+
+
+# Helper function to read some data from a file descriptor, ignoring EAGAIN errors
+def read_async(fd):
+    try:
+        return fd.read()
+    except IOError as exc:
+        if exc.errno != errno.EAGAIN:
+            raise exc
+        else:
+            return b''
+
+
 class FunctionRunnable(QRunnable):
     def __init__(self, func: Callable, args: List, kwargs: Dict):
         super().__init__()
@@ -91,7 +113,7 @@ class FunctionRunnable(QRunnable):
         self._args = args
         self._kwargs = kwargs
 
-    def run(self):
+    def run_(self):
         success = False
         self.signals.data.emit("-" * 20)
         self.signals.data.emit(
@@ -112,6 +134,78 @@ class FunctionRunnable(QRunnable):
         finally:
             self.signals.finished.emit(self._func.__name__, success)
 
+    def run(self):
+        tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
+        tmp.write(textwrap.dedent(
+            f"""\
+                from {self._func.__ui_module__} import {self._func.__name__}
+                response = {self._func.__name__}({stringify_args(self._args)}{', ' if self._args else ''}{stringify_kwargs(self._kwargs)})
+                if response is not None:
+                    print(response)
+            """
+        ))
+        tmp.close()
+        self.signals.data.emit("-" * 20)
+        options = dict(capture=True, capture_stderr=True, asynchronous=True, buffered=False, input=True)
+        try:
+            with ExternalCommand(f"{sys.executable} {tmp.name}", **options) as cmd:
+                make_async(cmd.stdout)
+                make_async(cmd.stderr)
+
+                while True:
+                    # Wait for data to become available
+                    out, *_ = select.select([cmd.stdout, cmd.stderr], [], [])
+
+                    # Try reading some data from each
+                    line_out = read_async(cmd.stdout)
+                    line_err = read_async(cmd.stderr)
+
+                    if line_out:
+                        self.signals.data.emit(line_out.decode(cmd.encoding).rstrip())
+                    if line_err:
+                        self.signals.data.emit(line_err.decode(cmd.encoding).rstrip())
+
+                    if cmd.subprocess.poll() is not None:
+                        break
+
+                # while cmd.is_running:
+                #     if cmd.stdout.readable():
+                #         out = cmd.stdout.readline().decode(cmd.encoding)
+                #         self.signals.data.emit(out.rstrip())
+                #     if cmd.stderr.readable():
+                #         err = cmd.stderr.readline().decode(cmd.encoding)
+                #         self.signals.data.emit(err.rstrip())
+
+                # for out, err in itertools.zip_longest(
+                #         # iter(lambda: cmd.stdout.readline().decode(cmd.encoding), u''),
+                #         # iter(lambda: cmd.stderr.readline().decode(cmd.encoding), u'')
+                #         iter(lambda: cmd.stdout.readline().decode(cmd.encoding), u''),
+                #         iter(lambda: cmd.stderr.readline().decode(cmd.encoding), u'')
+                # ):
+                #     out and self.signals.data.emit(out.rstrip())
+                #     err and self.signals.data.emit(err.rstrip())
+
+            cmd.wait()
+
+        except ExternalCommandFailed as exc:
+            # self._console_panel.append(cmd.error_message)
+            # This error message is also available in the decoded_stderr.
+            self.signals.error.emit(exc)
+
+        # if out := cmd.decoded_stdout:
+        #     self._console_panel.append(out)
+        if err := cmd.decoded_stderr:
+            self.signals.data.emit(err)
+
+        if cmd.is_finished:
+            if cmd.failed:
+                self.signals.finished.emit(self._func.__name__, False)
+            else:
+                self.signals.finished.emit(self._func.__name__, True)
+
+        # print(f"{tmp.name = }")
+        os.unlink(tmp.name)
+
     def start(self):
         QThreadPool.globalInstance().start(self)
 
@@ -189,6 +283,8 @@ class DynamicButton(QWidget):
         if final_stretch:
             layout.addStretch()
 
+        self.setToolTip(func.__doc__)
+
     @property
     def function(self) -> Callable:
         return self._function
@@ -206,7 +302,8 @@ class ArgumentsPanel(QGroupBox):
         super().__init__(button.label)
 
         self._button = button
-        self._args_fields = []
+        self._ui_args = ui_args
+        self._args_fields = {}
         self._kwargs_fields = {}
 
         # The arguments panel is a Widget with an input text field for each of the arguments.
@@ -218,8 +315,16 @@ class ArgumentsPanel(QGroupBox):
             input_field = QLineEdit()
             input_field.setObjectName(name)
             input_field.setPlaceholderText(str(arg.default) if arg.default else "")
+            if arg.annotation is not None:
+                input_field.setToolTip(f"The expected type is {arg.annotation.__name__}.")
+            else:
+                input_field.setToolTip("No type has been specified..")
+            if arg.annotation is int:
+                input_field.setValidator(QIntValidator())
+            elif arg.annotation is float:
+                input_field.setValidator(QDoubleValidator())
             if arg.kind == ArgumentKind.POSITIONAL_ONLY:
-                self._args_fields.append(input_field)
+                self._args_fields[name] = input_field
             elif arg.kind in [ArgumentKind.POSITIONAL_OR_KEYWORD, ArgumentKind.KEYWORD_ONLY]:
                 self._kwargs_fields[name] = input_field
             else:
@@ -242,17 +347,24 @@ class ArgumentsPanel(QGroupBox):
     @property
     def args(self):
         return [
-            arg.displayText()
-            for arg in self._args_fields
+            self._cast_arg(name, arg.displayText())
+            for name, arg in self._args_fields.items()
         ]
 
     @property
     def kwargs(self):
         return {
-            name: arg.displayText() or arg.placeholderText()
+            name: self._cast_arg(name, arg.displayText() or arg.placeholderText())
             for name, arg in self._kwargs_fields.items()
         }
 
+    def _cast_arg(self, name: str, value: Any):
+        arg = self._ui_args[name]
+        try:
+            return arg.annotation(value)
+        except (ValueError, TypeError):
+            return value
+
 
 class View(QMainWindow):
     def __init__(self):
@@ -262,7 +374,7 @@ class View(QMainWindow):
         self._buttons = []
         self.function_thread: FunctionRunnable
 
-        self.setWindowTitle("Contingency GUI")
+        self.setWindowTitle("Contingency GUI")  # TODO: set proper name of the GUI
 
         # self.setGeometry(300, 300, 500, 200)
 
@@ -331,45 +443,6 @@ class View(QMainWindow):
         if response := self._kernel.run_snippet(snippet):
             self.function_output(response)
 
-    def run_function_in_process(self, func: Callable, args: List, kwargs: Dict):
-        tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
-        tmp.write(textwrap.dedent(
-            f"""\
-                from {func.__ui_module__} import {func.__name__}
-                response = {func.__name__}({stringify_args(args)}{', ' if args else ''}{stringify_kwargs(kwargs)})
-                if response is not None:
-                    print(response)
-            """
-        ))
-        tmp.close()
-        self._console_panel.append("-" * 20)
-        cmd = ExternalCommand(
-            f"{sys.executable} {tmp.name}",
-            capture=True, capture_stderr=True, asynchronous=True
-        )
-        try:
-            cmd.start()
-            cmd.wait()
-        except ExternalCommandFailed as exc:
-            # self._console_panel.append(cmd.error_message)
-            # This error message is also available in the decoded_stderr.
-            ...
-
-        if out := cmd.decoded_stdout:
-            print(f"{out = }")
-            self._console_panel.append(out)
-        if err := cmd.decoded_stderr:
-            self._console_panel.append(err)
-
-        if cmd.is_finished:
-            if cmd.failed:
-                self._console_panel.append(f"function '{func.__name__}' finished execution.")
-            else:
-                self._console_panel.append(f"function '{func.__name__}' failed with an error.")
-
-        # print(f"{tmp.name = }")
-        os.unlink(tmp.name)
-
     def add_function_button(self, func: Callable):
 
         button = DynamicButton(func.__name__, func)
@@ -389,7 +462,7 @@ class View(QMainWindow):
 
         args_panel = ArgumentsPanel(button, ui_args)
         args_panel.run_button.clicked.connect(
-            lambda checked: self.run_function_in_process(args_panel.function, args_panel.args, args_panel.kwargs)
+            lambda checked: self.run_function(args_panel.function, args_panel.args, args_panel.kwargs)
         )
 
         if self._current_args_panel:
